from __future__ import annotations

import enum


def lookup_area(s: Area | str) -> Area:
    if isinstance(s, Area):
        # If it already is an Area object, we"re happy
        return s
    if isinstance(s, str):
        # If it is a "country code" string, we do a lookup
        if Area.has_code(s.upper()):
            return Area[s.upper()]

        # If it is a "direct code", we do a lookup
        for area in Area:
            if area.value == s:
                return area

    raise ValueError(f"Invalid country code, got {s=}")


class Area(enum.Enum):
    """ENUM containing 3 things about an Area: CODE, Meaning, Timezone."""

    def __new__(cls, *args, **kwds):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    # ignore the first param since it"s already set by __new__
    def __init__(self, _: str, meaning: str, tz: str):
        self._meaning = meaning
        self._tz = tz

    def __str__(self):
        return self.value

    @property
    def meaning(self):
        return self._meaning

    @property
    def tz(self):
        return self._tz

    @property
    def code(self):
        return self.value

    @classmethod
    def has_code(cls, code: str) -> bool:
        return code in cls.__members__

    # List taken directly from the API Docs
    # fmt: off
    DE_50HZ =        "10YDE-VE-------2", "50Hertz CA, DE(50HzT) BZA",                                  "Europe/Berlin",
    AL =             "10YAL-KESH-----5", "Albania, OST BZ / CA / MBA",                                 "Europe/Tirane",
    DE_AMPRION =     "10YDE-RWENET---I", "Amprion CA",                                                 "Europe/Berlin",
    AT =             "10YAT-APG------L", "Austria, APG BZ / CA / MBA",                                 "Europe/Vienna",
    BY =             "10Y1001A1001A51S", "Belarus BZ / CA / MBA",                                      "Europe/Minsk",
    BE =             "10YBE----------2", "Belgium, Elia BZ / CA / MBA",                                "Europe/Brussels",
    BA =             "10YBA-JPCC-----D", "Bosnia Herzegovina, NOS BiH BZ / CA / MBA",                  "Europe/Sarajevo",
    BG =             "10YCA-BULGARIA-R", "Bulgaria, ESO BZ / CA / MBA",                                "Europe/Sofia",
    CZ_DE_SK =       "10YDOM-CZ-DE-SKK", "BZ CZ+DE+SK BZ / BZA",                                       "Europe/Prague",
    HR =             "10YHR-HEP------M", "Croatia, HOPS BZ / CA / MBA",                                "Europe/Zagreb",
    CWE =            "10YDOM-REGION-1V", "CWE Region",                                                 "Europe/Brussels",
    CY =             "10YCY-1001A0003J", "Cyprus, Cyprus TSO BZ / CA / MBA",                           "Asia/Nicosia",
    CZ =             "10YCZ-CEPS-----N", "Czech Republic, CEPS BZ / CA/ MBA",                          "Europe/Prague",
    DE_AT_LU =       "10Y1001A1001A63L", "DE-AT-LU BZ",                                                "Europe/Berlin",
    DE_LU =          "10Y1001A1001A82H", "DE-LU BZ / MBA",                                             "Europe/Berlin",
    DK =             "10Y1001A1001A65H", "Denmark",                                                    "Europe/Copenhagen",
    DK_1 =           "10YDK-1--------W", "DK1 BZ / MBA",                                               "Europe/Copenhagen",
    DK_1_NO_1 =      "46Y000000000007M", "DK1 NO1 BZ",                                                 "Europe/Copenhagen",
    DK_2 =           "10YDK-2--------M", "DK2 BZ / MBA",                                               "Europe/Copenhagen",
    DK_CA =          "10Y1001A1001A796", "Denmark, Energinet CA",                                      "Europe/Copenhagen",
    EE =             "10Y1001A1001A39I", "Estonia, Elering BZ / CA / MBA",                             "Europe/Tallinn",
    FI =             "10YFI-1--------U", "Finland, Fingrid BZ / CA / MBA",                             "Europe/Helsinki",
    MK =             "10YMK-MEPSO----8", "Former Yugoslav Republic of Macedonia, MEPSO BZ / CA / MBA", "Europe/Skopje",
    FR =             "10YFR-RTE------C", "France, RTE BZ / CA / MBA",                                  "Europe/Paris",
    DE =             "10Y1001A1001A83F", "Germany",                                                    "Europe/Berlin"
    GR =             "10YGR-HTSO-----Y", "Greece, IPTO BZ / CA/ MBA",                                  "Europe/Athens",
    HU =             "10YHU-MAVIR----U", "Hungary, MAVIR CA / BZ / MBA",                               "Europe/Budapest",
    IS =             "IS",               "Iceland",                                                    "Atlantic/Reykjavik",
    IE_SEM =         "10Y1001A1001A59C", "Ireland (SEM) BZ / MBA",                                     "Europe/Dublin",
    IE =             "10YIE-1001A00010", "Ireland, EirGrid CA",                                        "Europe/Dublin",
    IT =             "10YIT-GRTN-----B", "Italy, IT CA / MBA",                                         "Europe/Rome",
    IT_SACO_AC =     "10Y1001A1001A885", "Italy_Saco_AC",                                              "Europe/Rome",
    IT_CALA =        "10Y1001C--00096J", "IT-Calabria BZ",                                             "Europe/Rome",
    IT_SACO_DC =     "10Y1001A1001A893", "Italy_Saco_DC",                                              "Europe/Rome",
    IT_BRNN =        "10Y1001A1001A699", "IT-Brindisi BZ",                                             "Europe/Rome",
    IT_CNOR =        "10Y1001A1001A70O", "IT-Centre-North BZ",                                         "Europe/Rome",
    IT_CSUD =        "10Y1001A1001A71M", "IT-Centre-South BZ",                                         "Europe/Rome",
    IT_FOGN =        "10Y1001A1001A72K", "IT-Foggia BZ",                                               "Europe/Rome",
    IT_GR =          "10Y1001A1001A66F", "IT-GR BZ",                                                   "Europe/Rome",
    IT_MACRO_NORTH = "10Y1001A1001A84D", "IT-MACROZONE NORTH MBA",                                     "Europe/Rome",
    IT_MACRO_SOUTH = "10Y1001A1001A85B", "IT-MACROZONE SOUTH MBA",                                     "Europe/Rome",
    IT_MALTA =       "10Y1001A1001A877", "IT-Malta BZ",                                                "Europe/Rome",
    IT_NORD =        "10Y1001A1001A73I", "IT-North BZ",                                                "Europe/Rome",
    IT_NORD_AT =     "10Y1001A1001A80L", "IT-North-AT BZ",                                             "Europe/Rome",
    IT_NORD_CH =     "10Y1001A1001A68B", "IT-North-CH BZ",                                             "Europe/Rome",
    IT_NORD_FR =     "10Y1001A1001A81J", "IT-North-FR BZ",                                             "Europe/Rome",
    IT_NORD_SI =     "10Y1001A1001A67D", "IT-North-SI BZ",                                             "Europe/Rome",
    IT_PRGP =        "10Y1001A1001A76C", "IT-Priolo BZ",                                               "Europe/Rome",
    IT_ROSN =        "10Y1001A1001A77A", "IT-Rossano BZ",                                              "Europe/Rome",
    IT_SARD =        "10Y1001A1001A74G", "IT-Sardinia BZ",                                             "Europe/Rome",
    IT_SICI =        "10Y1001A1001A75E", "IT-Sicily BZ",                                               "Europe/Rome",
    IT_SUD =         "10Y1001A1001A788", "IT-South BZ",                                                "Europe/Rome",
    RU_KGD =         "10Y1001A1001A50U", "Kaliningrad BZ / CA / MBA",                                  "Europe/Kaliningrad",
    LV =             "10YLV-1001A00074", "Latvia, AST BZ / CA / MBA",                                  "Europe/Riga",
    LT =             "10YLT-1001A0008Q", "Lithuania, Litgrid BZ / CA / MBA",                           "Europe/Vilnius",
    LU =             "10YLU-CEGEDEL-NQ", "Luxembourg, CREOS CA",                                       "Europe/Luxembourg",
    LU_BZN =         "10Y1001A1001A82H", "Luxembourg",                                                 "Europe/Luxembourg",
    MT =             "10Y1001A1001A93C", "Malta, Malta BZ / CA / MBA",                                 "Europe/Malta",
    ME =             "10YCS-CG-TSO---S", "Montenegro, CGES BZ / CA / MBA",                             "Europe/Podgorica",
    GB =             "10YGB----------A", "National Grid BZ / CA/ MBA",                                 "Europe/London",
    GE =             "10Y1001A1001B012", "Georgia",                                                    "Asia/Tbilisi",
    GB_IFA =         "10Y1001C--00098F", "GB(IFA) BZN",                                                "Europe/London",
    GB_IFA2 =        "17Y0000009369493", "GB(IFA2) BZ",                                                "Europe/London",
    GB_ELECLINK =    "11Y0-0000-0265-K", "GB(ElecLink) BZN",                                           "Europe/London",
    UK =             "10Y1001A1001A92E", "United Kingdom",                                             "Europe/London",
    NL =             "10YNL----------L", "Netherlands, TenneT NL BZ / CA/ MBA",                        "Europe/Amsterdam",
    NO_1 =           "10YNO-1--------2", "NO1 BZ / MBA",                                               "Europe/Oslo",
    NO_1A =          "10Y1001A1001A64J", "NO1 A BZ",                                                   "Europe/Oslo",
    NO_2 =           "10YNO-2--------T", "NO2 BZ / MBA",                                               "Europe/Oslo",
    NO_2_NSL =       "50Y0JVU59B4JWQCU", "NO2 NSL BZ / MBA",                                           "Europe/Oslo",
    NO_2A =          "10Y1001C--001219", "NO2 A BZ",                                                   "Europe/Oslo",
    NO_3 =           "10YNO-3--------J", "NO3 BZ / MBA",                                               "Europe/Oslo",
    NO_4 =           "10YNO-4--------9", "NO4 BZ / MBA",                                               "Europe/Oslo",
    NO_5 =           "10Y1001A1001A48H", "NO5 BZ / MBA",                                               "Europe/Oslo",
    NO =             "10YNO-0--------C", "Norway, Norway MBA, Stattnet CA",                            "Europe/Oslo",
    PL_CZ =          "10YDOM-1001A082L", "PL-CZ BZA / CA",                                             "Europe/Warsaw",
    PL =             "10YPL-AREA-----S", "Poland, PSE SA BZ / BZA / CA / MBA",                         "Europe/Warsaw",
    PT =             "10YPT-REN------W", "Portugal, REN BZ / CA / MBA",                                "Europe/Lisbon",
    MD =             "10Y1001A1001A990", "Republic of Moldova, Moldelectica BZ/CA/MBA",                "Europe/Chisinau",
    RO =             "10YRO-TEL------P", "Romania, Transelectrica BZ / CA/ MBA",                       "Europe/Bucharest",
    RU =             "10Y1001A1001A49F", "Russia BZ / CA / MBA",                                       "Europe/Moscow",
    SE_1 =           "10Y1001A1001A44P", "SE1 BZ / MBA",                                               "Europe/Stockholm",
    SE_2 =           "10Y1001A1001A45N", "SE2 BZ / MBA",                                               "Europe/Stockholm",
    SE_3 =           "10Y1001A1001A46L", "SE3 BZ / MBA",                                               "Europe/Stockholm",
    SE_4 =           "10Y1001A1001A47J", "SE4 BZ / MBA",                                               "Europe/Stockholm",
    RS =             "10YCS-SERBIATSOV", "Serbia, EMS BZ / CA / MBA",                                  "Europe/Belgrade",
    SK =             "10YSK-SEPS-----K", "Slovakia, SEPS BZ / CA / MBA",                               "Europe/Bratislava",
    SI =             "10YSI-ELES-----O", "Slovenia, ELES BZ / CA / MBA",                               "Europe/Ljubljana",
    GB_NIR =         "10Y1001A1001A016", "Northern Ireland, SONI CA",                                  "Europe/Belfast",
    ES =             "10YES-REE------0", "Spain, REE BZ / CA / MBA",                                   "Europe/Madrid",
    SE =             "10YSE-1--------K", "Sweden, Sweden MBA, SvK CA",                                 "Europe/Stockholm",
    CH =             "10YCH-SWISSGRIDZ", "Switzerland, Swissgrid BZ / CA / MBA",                       "Europe/Zurich",
    DE_TENNET =      "10YDE-EON------1", "TenneT GER CA",                                              "Europe/Berlin",
    DE_TRANSNET =    "10YDE-ENBW-----N", "TransnetBW CA",                                              "Europe/Berlin",
    TR =             "10YTR-TEIAS----W", "Turkey BZ / CA / MBA",                                       "Europe/Istanbul",
    UA =             "10Y1001C--00003F", "Ukraine, Ukraine BZ, MBA",                                   "Europe/Kiev",
    UA_DOBTPP =      "10Y1001A1001A869", "Ukraine-DobTPP CTA",                                         "Europe/Kiev",
    UA_BEI =         "10YUA-WEPS-----0", "Ukraine BEI CTA",                                            "Europe/Kiev",
    UA_IPS =         "10Y1001C--000182", "Ukraine IPS CTA",                                            "Europe/Kiev",
    XK =             "10Y1001C--00100H", "Kosovo/ XK CA / XK BZN",                                     "Europe/Rome",
    DE_AMP_LU =      "10Y1001C--00002H", "Amprion LU CA",                                              "Europe/Berlin",
    # fmt: on


# https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html#_psrtype
PSRTYPE_MAPPINGS = {
    "A03": "Mixed",
    "A04": "Generation",
    "A05": "Load",
    "B01": "Biomass",
    "B02": "Fossil Brown coal/Lignite",
    "B03": "Fossil Coal-derived gas",
    "B04": "Fossil Gas",
    "B05": "Fossil Hard coal",
    "B06": "Fossil Oil",
    "B07": "Fossil Oil shale",
    "B08": "Fossil Peat",
    "B09": "Geothermal",
    "B10": "Hydro Pumped Storage",
    "B11": "Hydro Run-of-river and poundage",
    "B12": "Hydro Water Reservoir",
    "B13": "Marine",
    "B14": "Nuclear",
    "B15": "Other renewable",
    "B16": "Solar",
    "B17": "Waste",
    "B18": "Wind Offshore",
    "B19": "Wind Onshore",
    "B20": "Other",
    "B21": "AC Link",
    "B22": "DC Link",
    "B23": "Substation",
    "B24": "Transformer",
    "B25": "Energy storage",
}

# https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html#_docstatus
DOCSTATUS = {
    "A01": "Intermediate",
    "A02": "Final",
    "A05": "Active",
    "A09": "Cancelled",
    "A13": "Withdrawn",
    "X01": "Estimated",
}

# https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html#_businesstype
BSNTYPE = {
    "A01": "Production",
    "A04": "Consumption",
    "A14": "Aggregated energy data",
    "A19": "Balance energy deviation",
    "A25": "General Capacity Information",
    "A29": "Already allocated capacity (AAC)",
    "A43": "Requested capacity (without price)",
    "A46": "System Operator redispatching",
    "A53": "Planned maintenance",
    "A54": "Unplanned outage",
    "A60": "Minimum possible",
    "A61": "Maximum possible",
    "A85": "Internal redispatch",
    "A91": "Positive forecast margin (if installed capacity > load forecast)",
    "A92": "Negative forecast margin (if load forecast > installed capacity)",
    "A93": "Wind generation",
    "A94": "Solar generation",
    "A95": "Frequency containment reserve",
    "A96": "Automatic frequency restoration reserve",
    "A97": "Manual frequency restoration reserve",
    "A98": "Replacement reserve",
    "B01": "Interconnector network evolution",
    "B02": "Interconnector network dismantling",
    "B03": "Counter trade",
    "B04": "Congestion costs",
    "B05": "Capacity allocated (including price)",
    "B07": "Auction revenue",
    "B08": "Total nominated capacity",
    "B09": "Net position",
    "B10": "Congestion income",
    "B11": "Production unit",
    "B33": "Area Control Error",
    "B74": "Offer",
    "B75": "Need",
    "B95": "Procured capacity",
    "C22": "Shared Balancing Reserve Capacity",
    "C23": "Share of reserve capacity",
    "C24": "Actual reserve capacity",
}

# https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html#_contract_marketagreement_type_type_marketagreement_type
MARKETAGREEMENTTYPE = {
    "A01": "Daily",
    "A02": "Weekly",
    "A03": "Monthly",
    "A04": "Yearly",
    "A05": "Total",
    "A06": "Long term",
    "A07": "Intraday",
    "A13": "Hourly",
}

# https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html#_documenttype
DOCUMENTTYPE = {
    "A09": "Finalised schedule",
    "A11": "Aggregated energy data report",
    "A15": "Acquiring system operator reserve schedule",
    "A24": "Bid document",
    "A25": "Allocation result document",
    "A26": "Capacity document",
    "A31": "Agreed capacity",
    "A38": "Reserve allocation result document",
    "A44": "Price Document",
    "A61": "Estimated Net Transfer Capacity",
    "A63": "Redispatch notice",
    "A65": "System total load",
    "A68": "Installed generation per type",
    "A69": "Wind and solar forecast",
    "A70": "Load forecast margin",
    "A71": "Generation forecast",
    "A72": "Reservoir filling information",
    "A73": "Actual generation",
    "A74": "Wind and solar generation",
    "A75": "Actual generation per type",
    "A76": "Load unavailability",
    "A77": "Production unavailability",
    "A78": "Transmission unavailability",
    "A79": "Offshore grid infrastructure unavailability",
    "A80": "Generation unavailability",
    "A81": "Contracted reserves",
    "A82": "Accepted offers",
    "A83": "Activated balancing quantities",
    "A84": "Activated balancing prices",
    "A85": "Imbalance prices",
    "A86": "Imbalance volume",
    "A87": "Financial situation",
    "A88": "Cross border balancing",
    "A89": "Contracted reserve prices",
    "A90": "Interconnection network expansion",
    "A91": "Counter trade notice",
    "A92": "Congestion costs",
    "A93": "DC link capacity",
    "A94": "Non EU allocations",
    "A95": "Configuration document",
    "B11": "Flow-based allocations",
    "B17": "Aggregated netted external TSO schedule document",
    "B45": "Bid Availability Document",
}

# https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html#_processtype
PROCESSTYPE = {
    "A01": "Day ahead",
    "A02": "Intra day incremental",
    "A16": "Realised",
    "A18": "Intraday total",
    "A31": "Week ahead",
    "A32": "Month ahead",
    "A33": "Year ahead",
    "A39": "Synchronisation process",
    "A40": "Intraday process",
    "A46": "Replacement reserve",
    "A47": "Manual frequency restoration reserve",
    "A51": "Automatic frequency restoration reserve",
    "A52": "Frequency containment reserve",
    "A56": "Frequency restoration reserve",
    "A60": "Scheduled activation mFRR",
    "A61": "Direct activation mFRR",
    "A67": "Central Selection aFRR",
    "A68": "Local Selection aFRR",
}

# neighbouring bidding zones that have cross_border flows
NEIGHBOURS = {
    "BE": ["NL", "DE_AT_LU", "FR", "GB", "DE_LU"],
    "NL": ["BE", "DE_AT_LU", "DE_LU", "GB", "NO_2", "DK_1"],
    "DE_AT_LU": ["BE", "CH", "CZ", "DK_1", "DK_2", "FR", "IT_NORD", "IT_NORD_AT", "NL", "PL", "SE_4", "SI"],
    "FR": ["BE", "CH", "DE_AT_LU", "DE_LU", "ES", "GB", "IT_NORD", "IT_NORD_FR"],
    "CH": ["AT", "DE_AT_LU", "DE_LU", "FR", "IT_NORD", "IT_NORD_CH"],
    "AT": ["CH", "CZ", "DE_LU", "HU", "IT_NORD", "SI"],
    "CZ": ["AT", "DE_AT_LU", "DE_LU", "PL", "SK"],
    "GB": ["BE", "FR", "IE_SEM", "NL", "NO_2", "DK_1"],
    "NO_2": ["DE_LU", "DK_1", "NL", "NO_1", "NO_5", "GB"],
    "HU": ["AT", "HR", "RO", "RS", "SI", "SK", "UA"],
    "IT_NORD": ["CH", "DE_AT_LU", "FR", "SI", "AT", "IT_CNOR"],
    "ES": ["FR", "PT"],
    "SI": ["AT", "DE_AT_LU", "HR", "IT_NORD"],
    "RS": ["AL", "BA", "BG", "HR", "HU", "ME", "MK", "RO"],
    "PL": ["CZ", "DE_AT_LU", "DE_LU", "LT", "SE_4", "SK", "UA"],
    "ME": ["AL", "BA", "RS"],
    "DK_1": ["DE_AT_LU", "DE_LU", "DK_2", "NO_2", "SE_3", "NL", "GB"],
    "RO": ["BG", "HU", "RS", "UA"],
    "LT": ["BY", "LV", "PL", "RU_KGD", "SE_4"],
    "BG": ["GR", "MK", "RO", "RS", "TR"],
    "SE_3": ["DK_1", "FI", "NO_1", "SE_2", "SE_4"],
    "LV": ["EE", "LT", "RU"],
    "IE_SEM": ["GB"],
    "BA": ["HR", "ME", "RS"],
    "NO_1": ["NO_2", "NO_3", "NO_5", "SE_3"],
    "SE_4": ["DE_AT_LU", "DE_LU", "DK_2", "LT", "PL", "SE_3"],
    "NO_5": ["NO_1", "NO_2", "NO_3"],
    "SK": ["CZ", "HU", "PL", "UA"],
    "EE": ["FI", "LV", "RU"],
    "DK_2": ["DE_AT_LU", "DE_LU", "DK_1", "SE_4"],
    "FI": ["EE", "NO_4", "RU", "SE_1", "SE_3"],
    "NO_4": ["SE_2", "FI", "NO_3", "SE_1"],
    "SE_1": ["FI", "NO_4", "SE_2"],
    "SE_2": ["NO_3", "NO_4", "SE_1", "SE_3"],
    "DE_LU": ["AT", "BE", "CH", "CZ", "DK_1", "DK_2", "FR", "NO_2", "NL", "PL", "SE_4"],
    "MK": ["BG", "GR", "RS"],
    "PT": ["ES"],
    "GR": ["AL", "BG", "IT_BRNN", "IT_GR", "MK", "TR"],
    "NO_3": ["NO_1", "NO_4", "NO_5", "SE_2"],
    "IT": ["AT", "FR", "GR", "MT", "ME", "SI", "CH"],
    "IT_BRNN": ["GR", "IT_SUD"],
    "IT_SUD": ["IT_BRNN", "IT_CSUD", "IT_FOGN", "IT_ROSN", "IT_CALA"],
    "IT_FOGN": ["IT_SUD"],
    "IT_ROSN": ["IT_SICI", "IT_SUD"],
    "IT_CSUD": ["IT_CNOR", "IT_SARD", "IT_SUD"],
    "IT_CNOR": ["IT_NORD", "IT_CSUD", "IT_SARD"],
    "IT_SARD": ["IT_CNOR", "IT_CSUD"],
    "IT_SICI": ["IT_CALA", "IT_ROSN", "MT"],
    "IT_CALA": ["IT_SICI", "IT_SUD"],
    "MT": ["IT_SICI"],
    "HR": ["BA", "HU", "RS", "SI"],
}



import httpx
import pytest

from denki_client.exceptions import (
    InvalidBusinessParameterError,
    InvalidPSRTypeError,
    NoMatchingDataError,
    PaginationError,
    raise_response_error,
)


@pytest.fixture(scope="session")
def mock_request():
    """Fixture to create a mock request object for responses."""
    return httpx.Request("GET", "https://example.com")


def test_successful_response(mock_request):
    """Test that a successful response is returned unchanged."""
    response = httpx.Response(
        200, content=b"<text>Valid data</text>", headers={"content-type": "application/xml"}, request=mock_request
    )
    result = raise_response_error(response)
    assert result == response


def test_no_matching_data_error_from_status(mock_request):
    """Test that NoMatchingDataError is raised when the error is in the status."""
    response = httpx.Response(400, content=b"<text>No matching data found</text>", request=mock_request)
    with pytest.raises(NoMatchingDataError):
        raise_response_error(response)


def test_no_matching_data_error_from_content(mock_request):
    """Test that NoMatchingDataError is raised when the error is in the content with status 200."""
    response = httpx.Response(
        200,
        content=b"<text>No matching data found</text>",
        headers={"content-type": "application/xml"},
        request=mock_request,
    )
    with pytest.raises(NoMatchingDataError):
        raise_response_error(response)


def test_invalid_business_parameter_error(mock_request):
    """Test that InvalidBusinessParameterError is raised for the corresponding error."""
    response = httpx.Response(
        400, content=b"<text>check you request against dependency tables</text>", request=mock_request
    )
    with pytest.raises(InvalidBusinessParameterError):
        raise_response_error(response)


def test_invalid_psr_type_error(mock_request):
    """Test that InvalidPSRTypeError is raised for the corresponding error."""
    response = httpx.Response(400, content=b"<text>is not valid for this area</text>", request=mock_request)
    with pytest.raises(InvalidPSRTypeError):
        raise_response_error(response)


def test_pagination_error_first_case(mock_request):
    """Test that PaginationError is raised for the first case with correct message."""
    content = b"<text>amount of requested data exceeds allowed limit of 500 items. Requested 600</text>"
    response = httpx.Response(400, content=content, request=mock_request)
    with pytest.raises(PaginationError) as exc_info:
        raise_response_error(response)
        assert "500" in str(exc_info.value)
        assert "600" in str(exc_info.value)


def test_pagination_error_second_case(mock_request):
    """Test that PaginationError is raised for the second case with correct message."""
    content = b"<text>requested data to be gathered via the offset parameter exceeds the allowed limit (500 items). Requested 600</text>"
    response = httpx.Response(400, content=content, request=mock_request)
    with pytest.raises(PaginationError) as exc_info:
        raise_response_error(response)
        assert "500" in str(exc_info.value)
        assert "600" in str(exc_info.value)


def test_other_http_error(mock_request):
    """Test that other HTTP errors are propagated correctly."""
    response = httpx.Response(500, content=b"<text>Internal server error</text>", request=mock_request)
    with pytest.raises(httpx.HTTPError):
        raise_response_error(response)


def test_non_xml_content_with_200_status(mock_request):
    """Test that non-XML content with 200 status is returned unchanged."""
    response = httpx.Response(
        200, content=b"Some non-XML data", headers={"content-type": "text/plain"}, request=mock_request
    )
    result = raise_response_error(response)
    assert result == response


def test_zip_content_with_200_status(mock_request):
    """Test that zip content with 200 status is returned unchanged."""
    response = httpx.Response(
        200, content=b"ZIP data", headers={"content-type": "application/zip"}, request=mock_request
    )
    result = raise_response_error(response)
    assert result == response


def test_empty_response(mock_request):
    """Test that an empty response raises HTTPError."""
    response = httpx.Response(400, content=b"", request=mock_request)
    with pytest.raises(httpx.HTTPError):
        raise_response_error(response)


def test_response_without_text_tag(mock_request):
    """Test that a response without text tag raises HTTPError."""
    response = httpx.Response(400, content=b"<html><body>Error</body></html>", request=mock_request)
    with pytest.raises(httpx.HTTPError):
        raise_response_error(response)



import re
from datetime import UTC, datetime, tzinfo

from dateutil.relativedelta import relativedelta
from pytz import timezone

from .exceptions import ParseError, TzNaiveError


def parse_datetime(date: datetime | str, target_tz: tzinfo | str = UTC) -> datetime:
    """Parse Datetime with timezone.

    :param datetime | str date: datetime to parse. If datetime, must be tz-aware
    :param tzinfo target_tz: target timezone, defaults to UTC
    :raises TzNaiveError:
    :return datetime: Datetime in format yyyyMMddHHmm
    """
    if isinstance(target_tz, str):
        target_tz = timezone(target_tz)
    if isinstance(date, datetime):
        if date.tzinfo is None:
            raise TzNaiveError
        date = date.astimezone(target_tz)
    elif isinstance(date, str):
        date = datetime.fromisoformat(date).astimezone(target_tz)
    return date


def parse_freq(freq: relativedelta | str) -> relativedelta:
    """Parse a time string e.g. (2h13m) into a relativedelta object.

    :param relativedelta | str freq: A relativedelta or a string identifying a duration. (eg. 2h13m)
    :return relativedelta:
    """
    if isinstance(freq, str):
        regex = re.compile(
            r"^((?P<years>[\.\d]+?)y)?((?P<months>[\.\d]+?)mo)?((?P<days>[\.\d]+?)d)?((?P<hours>[\.\d]+?)h)?((?P<minutes>[\.\d]+?)m)?((?P<seconds>[\.\d]+?)s)?$"
        )
        parts = regex.match(freq)
        if parts is None:
            raise ParseError(
                f"Could not parse any time information from '{freq}'. Examples of valid strings: '8h', '2d8h5m20s','2m4s', '1y2mo'"
            )
        time_params = {name: float(param) for name, param in parts.groupdict().items() if param}
    return relativedelta(**time_params)



def hello_from_bin() -> str:
    """Returns a str from Rust backend."""

def parse_timeseries_generic(
    xml_text: str, labels: list[str], metadata: list[str], period_name: str
) -> dict[str, list[str]]:
    """Parses XML-formatted string to a Python dictionnary.

    :param str xml_text: XML-formatted string
    :param list[str] labels: list of XML tags to retreive inside a `period_name` tag
    :param list[str] metadata: list of XML tags to retreive between a `TimeSeries` and a `period_name` tag
    :param str period_name: usually 'period'
    :return dict[str, list[str]]:
    """



from datetime import UTC, datetime

import pytest
from dateutil.relativedelta import relativedelta

from denki_client.utils import yield_date_range


@pytest.mark.parametrize(
    "start, end, freq, expected",
    [
        (
            datetime(2023, 1, 1, tzinfo=UTC),
            datetime(2025, 1, 2, tzinfo=UTC),
            relativedelta(years=1),
            [
                (datetime(2023, 1, 1, tzinfo=UTC), datetime(2024, 1, 1, tzinfo=UTC)),
                (datetime(2024, 1, 1, tzinfo=UTC), datetime(2025, 1, 1, tzinfo=UTC)),
                (datetime(2025, 1, 1, tzinfo=UTC), datetime(2025, 1, 2, tzinfo=UTC)),
            ],
        ),
        (
            datetime(2023, 1, 1, tzinfo=UTC),
            datetime(2023, 3, 2, tzinfo=UTC),
            relativedelta(months=1),
            [
                (datetime(2023, 1, 1, tzinfo=UTC), datetime(2023, 2, 1, tzinfo=UTC)),
                (datetime(2023, 2, 1, tzinfo=UTC), datetime(2023, 3, 1, tzinfo=UTC)),
                (datetime(2023, 3, 1, tzinfo=UTC), datetime(2023, 3, 2, tzinfo=UTC)),
            ],
        ),
        (
            datetime(2023, 1, 1, tzinfo=UTC),
            datetime(2023, 1, 3, 23, tzinfo=UTC),
            relativedelta(days=1),
            [
                (datetime(2023, 1, 1, tzinfo=UTC), datetime(2023, 1, 2, tzinfo=UTC)),
                (datetime(2023, 1, 2, tzinfo=UTC), datetime(2023, 1, 3, tzinfo=UTC)),
                (datetime(2023, 1, 3, tzinfo=UTC), datetime(2023, 1, 3, 23, tzinfo=UTC)),
            ],
        ),
        (
            datetime(2023, 1, 1, tzinfo=UTC),
            datetime(2023, 1, 1, 2, 45, tzinfo=UTC),
            relativedelta(hours=1),
            [
                (datetime(2023, 1, 1, tzinfo=UTC), datetime(2023, 1, 1, 1, tzinfo=UTC)),
                (datetime(2023, 1, 1, 1, tzinfo=UTC), datetime(2023, 1, 1, 2, tzinfo=UTC)),
                (datetime(2023, 1, 1, 2, tzinfo=UTC), datetime(2023, 1, 1, 2, 45, tzinfo=UTC)),
            ],
        ),
        (
            datetime(2023, 1, 1, tzinfo=UTC),
            datetime(2023, 1, 1, 0, 2, 45, tzinfo=UTC),
            relativedelta(minutes=1),
            [
                (datetime(2023, 1, 1, tzinfo=UTC), datetime(2023, 1, 1, 0, 1, tzinfo=UTC)),
                (datetime(2023, 1, 1, 0, 1, tzinfo=UTC), datetime(2023, 1, 1, 0, 2, tzinfo=UTC)),
                (datetime(2023, 1, 1, 0, 2, tzinfo=UTC), datetime(2023, 1, 1, 0, 2, 45, tzinfo=UTC)),
            ],
        ),
        (
            datetime(2023, 1, 1, tzinfo=UTC),
            datetime(2023, 1, 1, 0, 0, 2, tzinfo=UTC),
            relativedelta(seconds=1),
            [
                (datetime(2023, 1, 1, tzinfo=UTC), datetime(2023, 1, 1, 0, 0, 1, tzinfo=UTC)),
                (datetime(2023, 1, 1, 0, 0, 1, tzinfo=UTC), datetime(2023, 1, 1, 0, 0, 2, tzinfo=UTC)),
            ],
        ),
    ],
)
def test_yield_date_range(start, end, freq, expected):
    result = list(yield_date_range(start, end, freq))
    assert result == expected



import asyncio
import os

import narwhals as nw
import pytest

from denki_client.entsoe import EntsoeClient


@pytest.fixture
def client():
    return EntsoeClient(os.environ["API_KEY_ENTSOE"], backend="polars")


def test_query_day_ahead_prices(client: EntsoeClient):
    df = asyncio.run(client.query_day_ahead_prices("FR", start="20250101", end="20250103"))
    assert isinstance(df, nw.DataFrame)


def test_query_activated_balancing_energy_prices(client: EntsoeClient):
    df = asyncio.run(
        client.query_activated_balancing_energy_prices(
            "FR",
            "A16",
            "A95",
            start="20250101",
            end="20250103",
        )
    )
    assert isinstance(df, nw.DataFrame)



import re

import httpx


class PaginationError(Exception):
    pass


class NoMatchingDataError(Exception):
    pass


class InvalidPSRTypeError(Exception):
    pass


class InvalidBusinessParameterError(Exception):
    pass


class InvalidParameterError(Exception):
    pass


class TzNaiveError(Exception):
    pass


class ParseError(Exception):
    pass


def raise_response_error(response: httpx.Response):
    """Raises correct error from Entsoe server response.

    :param httpx.Response response:
    :raises NoMatchingDataError:
    :raises InvalidBusinessParameterError:
    :raises InvalidPSRTypeError:
    :raises PaginationError:
    :return httpx.Response:
    """
    try:
        response.raise_for_status()
    except httpx.HTTPError as e:
        error_text = ""
        text_match = re.search(r"<text>(.*?)</text>", response.text, re.DOTALL)
        if text_match:
            error_text = text_match.group(1).strip()

        if error_text:
            if "No matching data found" in error_text:
                raise NoMatchingDataError
            elif "check you request against dependency tables" in error_text:
                raise InvalidBusinessParameterError
            elif "is not valid for this area" in error_text:
                raise InvalidPSRTypeError
            elif "amount of requested data exceeds allowed limit" in error_text:
                numbers = [int(s) for s in error_text.split() if s.isdigit()]
                if len(numbers) >= 2:
                    allowed, requested = numbers[-2], numbers[-1]
                else:
                    allowed, requested = "unknown", "unknown"
                raise PaginationError(
                    f"The API is limited to {allowed} elements per request. This query requested {requested} documents."
                )
            elif "requested data to be gathered via the offset parameter exceeds the allowed limit" in error_text:
                numbers = [int(s) for s in error_text.split() if s.isdigit()]
                if len(numbers) >= 1:
                    allowed = numbers[0]
                    requested_match = re.search(r"Requested[^\d]*(\d+)", error_text)
                    requested = requested_match.group(1) if requested_match else "unknown"
                else:
                    allowed, requested = "unknown", "unknown"
                raise PaginationError(
                    f"The API is limited to {allowed} elements per request. This query requested {requested} documents."
                )
        raise e
    # ENTSO-e has changed their server to also respond with 200 if there is no data but all parameters are valid
    # this means we need to check the contents for this error even when status code 200 is returned
    # to prevent parsing the full response do a text matching instead of full parsing
    # also only do this when response type content is text and not for example a zip file.
    if response.headers.get("content-type", "") == "application/xml":
        if "No matching data found" in response.text:
            raise NoMatchingDataError
    return response



from datetime import UTC

import narwhals as nw

DAY_AHEAD_SCHEMA = nw.Schema(
    {
        "timestamp": nw.Datetime(time_zone=UTC),
        "price.amount": nw.Float64,
        "resolution": nw.Enum(["PT60M", "PT30M", "PT15M"]),
    }
)

ACTIVATED_BALANCING_ENERGY_PRICES_SCHEMA = nw.Schema(
    {
        "timestamp": nw.Datetime(time_zone=UTC),
        "activation_Price.amount": nw.Float64,
        "flowDirection.direction": nw.Enum(["A01", "A02"]),
        "businessType": nw.Enum(["A95", "A96", "A97", "A98"]),
        "resolution": nw.Enum(["PT60M", "PT30M", "PT15M"]),
    }
)



import logging
from datetime import datetime
from functools import wraps
from typing import Literal

import narwhals as nw
from dateutil.relativedelta import relativedelta

from denki_client.area import Area, lookup_area
from denki_client.parsers import parse_datetime, parse_freq

from .exceptions import NoMatchingDataError, PaginationError

logger = logging.getLogger(__name__)


def parse_inputs(func):
    """Parses function inputs.

    - country_code: `Area` | `str` -> `Area`
    - start: `datetime` | `str` -> `str`
    - end: `datetime` | `str` -> `str`
    """

    @wraps(func)
    async def parse_inputs_wrapper(
        self, country_code: Area | str, *args, start: datetime | str, end: datetime | str, **kwargs
    ):
        country_code = lookup_area(country_code)
        start = parse_datetime(start, country_code.tz)
        end = parse_datetime(end, country_code.tz)
        return await func(self, country_code, *args, start=start, end=end, **kwargs)

    return parse_inputs_wrapper


def paginated(func):
    """Catches a PaginationError, splits the requested period in two and tries
    again. Finally it concatenates the results.
    """

    @wraps(func)
    async def pagination_wrapper(*args, start: datetime, end: datetime, **kwargs):
        try:
            frame: nw.DataFrame | None = await func(*args, start=start, end=end, **kwargs)

        except PaginationError:
            pivot = start + (end - start) / 2
            frame1: nw.DataFrame | None = await pagination_wrapper(*args, start=start, end=pivot, **kwargs)
            frame2: nw.DataFrame | None = await pagination_wrapper(*args, start=pivot, end=end, **kwargs)

            if frame1 is None and frame2 is None:
                frame = None
            else:
                frame = nw.concat([frame1, frame2], how="diagonal")
        return frame

    return pagination_wrapper


def documents_limited(n: int = 100):
    """Deals with calls where you cannot query more than n documents at a
    time, by offsetting per `n` documents. Function needs `offset` kwarg.

    :param int n: maximum number of documents to query, defaults to 100.
    """

    def decorator(func):
        @wraps(func)
        async def documents_wrapper(*args, **kwargs):
            frames = []
            for _offset in range(0, 4800 + n, n):
                try:
                    frame: nw.DataFrame | None = await func(*args, offset=_offset, **kwargs)
                    if frame is not None:
                        frames.append(frame)

                except NoMatchingDataError:
                    logger.debug(f"NoMatchingDataError: for offset {_offset}")
                    break

            if frames == []:
                logger.debug("All the data returned are void")

            df = nw.concat(frames, how="diagonal")
            return df

        return documents_wrapper

    return decorator


def yield_date_range(start: datetime, end: datetime, freq: relativedelta):
    """Create a date_range iterator from `start` to `end` at given frequency.

    :param datetime start:
    :param datetime end:
    :param relativedelta freq:
    :yield datetime: _start
    :yield datetime: _end
    """

    _t0 = start
    while _t0 < end:
        _t1 = min(_t0 + freq, end)
        yield _t0, _t1
        _t0 = _t1


def split_query(freq: relativedelta | str):
    """Deals with calls where you cannot query more than a given frequency,
    by splitting the call up in blocks. Function needs `start` and `end` kwargs.

    :param relativedelta | str freq: split frequency compatible with `parse_freq`
    """
    freq: relativedelta = parse_freq(freq)

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, start: datetime, end: datetime, **kwargs):
            blocks = yield_date_range(start, end, freq)
            frames = []
            for _start, _end in blocks:
                try:
                    frame: nw.DataFrame | None = await func(*args, start=_start, end=_end, **kwargs)
                    if frame is not None:
                        frames.append(frame)

                except NoMatchingDataError:
                    logger.debug(f"NoMatchingDataError: between {_start} and {_end}")

            if frames == []:
                logger.debug("All the data returned are void")
                return None

            df = nw.concat(frames, how="diagonal")
            return df

        return wrapper

    return decorator


def inclusive(granularity: relativedelta | str, closed: Literal["both", "left", "right", "neither"]):
    """Truncate `start` and `end` arguments for calls.

    :param relativedelta | str granularity:
    :param Literal["both", "left", "right", "neither"] closed: where the interval is closed
    """
    granularity = parse_freq(granularity)
    granularity: relativedelta

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, start: datetime, end: datetime, **kwargs):
            match closed:
                case "both":
                    _start = start
                    _end = end
                case "left":
                    _start = start
                    _end = end - granularity
                case "right":
                    _start = start + granularity
                    _end = end
                case "neither":
                    _start = start + granularity
                    _end = end - granularity
            return await func(*args, start=_start, end=_end, **kwargs)

        return wrapper

    return decorator



import logging
from datetime import datetime
from types import ModuleType
from typing import Literal

import httpx
import narwhals as nw
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed

from denki_client._core import parse_timeseries_generic
from denki_client.area import Area
from denki_client.exceptions import raise_response_error
from denki_client.schemas import ACTIVATED_BALANCING_ENERGY_PRICES_SCHEMA, DAY_AHEAD_SCHEMA
from denki_client.utils import documents_limited, inclusive, parse_inputs, split_query


class EntsoeClient:
    def __init__(self, api_key: str, backend: ModuleType | nw.Implementation | str, **httpx_client_kwargs) -> None:
        """Client to ENTSO-e API.

        :param str api_key: API key obtained by creating an account on the website.
        :param ModuleType | Implementation | str backend: Narwhals's compatible backend.
        :param dict httpx_client_kwargs: Additional keyword arguments to pass to the httpx client.

        API doc: `https://documenter.getpostman.com/view/7009892/2s93JtP3F6`.
        """
        self.api_key = api_key
        self.base_url = "https://web-api.tp.entsoe.eu/api"
        self.session = httpx.AsyncClient(**httpx_client_kwargs)
        self.logger = logging.getLogger(__name__)
        self.backend = backend

    @retry(
        retry=retry_if_exception_type(httpx.ConnectError),
        stop=stop_after_attempt(3),
        wait=wait_fixed(2),
    )
    async def _base_request(self, params: dict, start_str: str, end_str: str) -> httpx.Response:
        """Base Request.

        :param dict params: parameters dictionnary. See documentation for more details.
        :param str start_str: Pattern yyyyMMddHHmm e.g. 201601010000. Considered timezone is the local one.
        :param str end_str: Pattern yyyyMMddHHmm e.g. 201601010000 Considered timezone is the local one.
        :return httpx.Response:
        """
        base_params = {
            "securityToken": self.api_key,
            "periodStart": start_str,
            "periodEnd": end_str,
        }
        params.update(base_params)
        params = {k: v for k, v in params.items() if v is not None}
        self.logger.debug(f"Request with {params=}")
        response = await self.session.get(self.base_url, params=params)
        raise_response_error(response)
        return response

    def _prepare_inputs(
        self, country_code: Area | str, start: datetime | str, end: datetime | str
    ) -> tuple[str, str, str]:
        if isinstance(country_code, str):
            raise TypeError(f"{type(country_code)=} instead of Area. Consider using the `parse_inputs` decorator.")

        if isinstance(start, str) or isinstance(end, str):
            raise TypeError(
                f"(type(start), type(end)) = ({type(start)}, {type(end)}) instead of (datetime, datetime). Consider using the `parse_inputs` decorator."
            )

        start_str = start.strftime("%Y%m%d%H%M")
        end_str = end.strftime("%Y%m%d%H%M")
        return country_code.code, start_str, end_str

    @parse_inputs
    @split_query("1y")
    @documents_limited(100)
    @inclusive("1d", "left")
    async def query_day_ahead_prices(
        self, country_code: Area | str, *, start: datetime | str, end: datetime | str, offset: int = 0
    ) -> nw.DataFrame | None:
        """Query day-ahead prices.

        API documentation: `https://documenter.getpostman.com/view/7009892/2s93JtP3F6#3b383df0-ada2-49fe-9a50-98b1bb201c6b`

        :param  Area | str country_code:
        :param datetime | str start: start of the query
        :param datetime | str end: end of the query
        :param int offset: defaults to 0
        :return nw.DataFrame | None: DataFrame with the following columns:
        - timestamp: in UTC
        - price.amount: in €/MWh
        - resolution
        """
        domain_code, start_str, end_str = self._prepare_inputs(country_code, start, end)

        params = {
            "documentType": "A44",
            "in_Domain": domain_code,
            "out_Domain": domain_code,
            "contract_MarketAgreement.type": "A01",
            "offset": offset,
        }
        response = await self._base_request(params, start_str, end_str)
        data = parse_timeseries_generic(
            response.text,
            ["price.amount"],
            [],
            "period",
        )
        if data == {}:
            return None
        df = nw.from_dict(data, DAY_AHEAD_SCHEMA, backend=self.backend)
        return df

    @parse_inputs
    @split_query("1y")
    async def query_activated_balancing_energy_prices(
        self,
        country_code: Area | str,
        process_type: Literal["A16", "A60", "A61", "A68", None],
        business_type: Literal["A95", "A96", "A97", "A98", None],
        *,
        start: datetime | str,
        end: datetime | str,
    ) -> nw.DataFrame | None:
        """Query activated balancing energy prices.

        API documentation: `https://documenter.getpostman.com/view/7009892/2s93JtP3F6#c301d91e-53ac-4aca-8e18-f29e9146c4a6`

        :param  Area | str country_code:
        :param Literal['A16', 'A60', 'A61', 'A68', None] process_type:
        - A16: Realised
        - A60: Scheduled activation mFRR
        - A61: Direct activation mFRR
        - A68: Local Selection aFRR
        - None: select all
        :param Literal['A95', 'A96', 'A97', 'A98', None] business_type:
        - A95: Frequency containment reserve
        - A96: Automatic frequency restoration reserve
        - A97: Manual frequency restoration reserve
        - A98: Replacement reserve
        - None: select all
        :param datetime | str start: start of the query
        :param datetime | str end: end of the query
        :return nw.DataFrame | None: DataFrame with the following columns:
        - timestamp: in UTC
        - activation_Price.amount: in €/MWh
        - flowDirection.direction: Up is A01 and Down is A02
        - businessType
        - resolution
        """
        domain_code, start_str, end_str = self._prepare_inputs(country_code, start, end)

        params = {
            "documentType": "A84",
            "processType": process_type,
            "controlArea_Domain": domain_code,
            "businessType": business_type,
            "psrType": None,
            "standardMarketProduct": None,
            "originalMarketProduct": None,
        }
        response = await self._base_request(params, start_str, end_str)
        data = parse_timeseries_generic(
            response.text,
            ["activation_Price.amount"],
            ["flowDirection.direction", "businessType"],
            "period",
        )
        if data == {}:
            return None
        df = nw.from_dict(data, ACTIVATED_BALANCING_ENERGY_PRICES_SCHEMA, backend=self.backend)
        return df
